// ####################################################
// ##  K7DUN                                         ##
// ##  CW-ELMER 3.0 rev 031 02/07/2020               ##
// ####################################################
#include <SoftwareSerial.h>
#include <stdlib.h>
#include <EEPROM.h>
//-------------------------------------------------------

#define EEP_QRZ     0                                    // EEPROM 00-15 = 16 chars (qrz string, 15 byte string + NULL)
#define EEP_WPM     16                                   // EEPROM 16-17 = 1 int (wpm int, 2 bytes)
#define EEP_ANP     18                                   // EEPROM 18-19 = int (anp int, 2 bytes)
#define EEP_PIT     20                                   // EEPROM 20-21 = int (pitch int, 2 bytes)
//#define EEP_ROM     22                                   // EEPROM 22-23 = NOT DEFINED AT THIS TIME
#define EEP_DIT     24                                   // EEPROM 24-25 = int (ditTime int, 2 bytes)
#define EEP_MID     26                                   // EEPROM 26-27 = int (midTime int, 2 bytes)
#define EEP_DAH     28                                   // EEPROM 28-29 = int (dahTime int, 2 bytes)

#define PIN_KEY     2                                    // Arduino Pin D2 for Morse straight key, to ground
#define PIN_SPK     4                                    // Arduino Pin D4 for Audio out to speaker, to ground

#define LCD_COL     16                                   // LCD maximum chars per row
#define LCD_ON      0x16                                 // LCD Code to turn LCD display on, no cursor
#define LCD_OC      0x18                                 // LCD Code to turn LCD display on, with cursor
#define LCD_LIT     0x11                                 // LCD Code to turn LCD backlight on
#define LCD_CLR     0x0C                                 // LCD Code to clear LCD display
#define LCD_RO0     0x80                                 // LCD Select LCD row 0 (upper), character 0 (leftmost)
#define LCD_RO1     0x94                                 // LCD Select LCD row 1 (lower), character 0 (leftmost)

#define ONESEC      1000                                 // 1000 miliseconds = 1 second for delays
#define TWOSEC      2000                                 // 2000 milliseconds = 2 seconds for delays
#define FIVSEC      5000                                 // 5000 milliseconds = 5 seconds for delays
#define CQ          99                                   // Sound "CQ" in Morse
//-------------------------------------------------------
char
  qrz[]       = "    WB6LAU     ",                       // Qrz buffer with default string (EEPROM)
  ver[]       = "CW-ELMER    3.0",                       // Ver buffer, not adjustable
  lcd0[LCD_COL],                                         // LCD ROW 0 (UPPER) buffer
  lcd1[LCD_COL],                                         // LCD ROW 1 (LOWER) buffer
  charLUT[]   = " ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789(/=:,)!;-'@._?$";                // Character lookup table
//-------------------------------------------------------
unsigned int
  wpm         = 15,                                      // Words-Per-Minute rate, default = 15 (EEPROM)
  ditTime     = 80,                                      // DIT timer, default = 80 mS (EEPROM)
  ditMin      = 9999,                                    // Shortest DIT measured for stats
  ditMax      = 0,                                       // Longest DIT measured for stats
  ditAcc      = 0,                                       // Accumulated DIT time
  ditCount    = 0,                                       // Count DITs detected
  dahTime     = 240,                                     // DAH timer, default = 240 mS (EEPROM)
  dahMin      = 9999,                                    // Shortest DAH measured for stats
  dahMax      = 0,                                       // Longest DAH measured for stats
  dahAcc      = 0,                                       // Accumulated DAH time
  dahCount    = 0,                                       // Count DAHs detected
  midTime     = 160,                                     // Threshold to determine if DIT or DAH detected (EEPROM)
  downTime    = 0,                                       // Key depressed timer (milliseconds)
  upTime      = 0,                                       // Key released timer (milliseconds)
  anp         = 0,                                       // Character set 0=alpha, 1=alpha+numer, 2=alpha+numer+punct (EEPROM)
  charIndex   = 0,                                       // Used for interpreting Morse code as keyed
  pitch       = 3,                                       // NOTE pitch poiner for pitchLUT, default = 3 (700Hz) (EEPROM)
  pitchLUT[5] =                                          // CWPitch frequency table :: pitchLUT[note]
    {400, 500, 600, 700, 800};
//-------------------------------------------------------
long
  downStart   = 0,                                       // Key depressed start time (reference)
  upStart     = 0;                                       // Key released start time (reference)
//-------------------------------------------------------
bool
  charDone    = false,                                   // Used in SendMode()
  toneDone    = false,                                   // Used in SendMode()
  smLoop      = true,                                    // Used in SendMode()
  sound       = true;                                    // false=sound disabled :: true=sound enabled, for development
//-------------------------------------------------------

SoftwareSerial lcd(0, 1);                                // Initialize LCD communications pins: RX = pin 0, TX = pin 1
//-------------------------------------------------------

// ####################################################
// ##  A R D U I N O   S E T U P                     ##
// ####################################################
void setup() {
  pinMode(PIN_KEY, INPUT_PULLUP);                        // Initialize KEY input pin on Arduino
  lcd.begin(2400); delay(100);                           // Initialize LCD serial port to 2400 baud
  Serial.begin(9600);                                    // Initialize USB serial port to 9600 baud
  lcd.write(LCD_CLR); delay(50);
  lcd.write(LCD_ON);
  lcd.write(LCD_LIT); delay(100);

  if(!digitalRead(PIN_KEY)) EepromInit();                // Initialize EEPROM for first-time use (key during power on)

  EepromRead();                                          // Retrieve saved values from EEPROM
  PrintLCD(qrz, ver); delay(ONESEC);

  if(!digitalRead(PIN_KEY)) {                            // If keypress detected turn sound OFF
    sound = false;
    strcpy(lcd1,"  SILENT MODE  ");
    PrintLCD(qrz, lcd1);
  }

  SoundQRZ(); delay(TWOSEC);
  lcd.write(LCD_CLR); delay(50);
}
//-------------------------------------------------------

// ####################################################
// ##  A R D U I N O   L O O P                       ##
// ##  (MAIN MENU)                                   ##
// ####################################################
void loop() {
  int indx = 1;
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

  strcpy(lcd0, "CW-Elmer   MENU");
  while(true) {                                          // Run menu loop until item selected
    switch(indx) {
      case 1: strcpy(lcd1, "    Sending    "); break;
      case 2: strcpy(lcd1, "   Receiving   "); break;
      case 3: strcpy(lcd1, "    Analyze    "); break;
      case 4: strcpy(lcd1, "   Exercises   "); break;
      case 5: strcpy(lcd1, "    Options    "); break;
    }
    PrintLCD(lcd0, lcd1);
    if(KeyDelay(TWOSEC)) break;                          // key pressed, exit from menu loop, pass indx to processing
    if(++indx > 5) indx = 1;                             // Otherwise increment indx and repeat menu loop
  }

  Click(); MarkLCD();                                    // Acknowledge selection
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
  lcd.write(LCD_CLR); delay(50);

  switch(indx) {                                         // Process function per indx value
    case 1: SendMode(); break;
    case 2: RecvMode(); break;
    case 3: Analyze(); break;
    case 4: Exercise(); break;
    case 5: OptionsMode(); break;
    default: break;                                      // If indx out of range do nothing (error capturing)
  }
}
//-------------------------------------------------------

// ####################################################
// ##  S E N D I N G   M O D E                       ##
// ####################################################
void SendMode() {
  strcpy(lcd0, " SENDING  MODE ");
  strcpy(lcd1, ver);
  PrintLCD(lcd0, lcd1);                                  // Print to LCD display and USB serial
  SoundCode(CQ);                                         // Sound CQ

  strcpy(lcd0, "KEY Morse chars");
  strcpy(lcd1, "Hold >2sec exit");
  PrintLCD(lcd0, lcd1); delay(TWOSEC);
  lcd0[0] = '\0';                                        // ERROR CORRECTION 3/10/20: Pre-clear lcd0 string
  lcd.write(LCD_CLR); delay(50);
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

  smLoop = true;
  while(smLoop) {                                        // Monitor KEY state in continuous loop
      digitalRead(PIN_KEY) ? KeyIsUp() : KeyIsDown();    // If PIN_KEY is high (open) run KeyIsUp(), if low (pressed) run KeyIsDown()
  }
}
//-------------------------------------------------------

void KeyIsDown() {
  sound ? tone(PIN_SPK, pitchLUT[pitch]) : DoNothing();      // Start audio tone, continue until key released

  upStart = 0;                                           // Reset upStart counter
  if(downStart == 0) downStart = millis();               // If downStart is 0, set to current millis()
  charDone = true;
  toneDone = true;
  if(charIndex == 0) charIndex = 1;

  if(millis() - downStart >= TWOSEC) {                   // If KEY depressed >2 seconds, exit and return to MAIN MENU
    noTone(PIN_SPK);                                     // Turn tone OFF
    smLoop = false;                                      // Set smLoop to false to force exit from loop
  }
}
//-------------------------------------------------------

void KeyIsUp() {
  noTone(PIN_SPK);

  if(upStart == 0) upStart = millis();                   // Set upStart timer
  upTime = millis() - upStart;                           // Measure time KEY is up

  if(downStart > 0) {
    downTime = millis() - downStart;                     // Measure time KEY was down
    downStart = 0;                                       // Reset downStart timer
  }

  if(toneDone) ShiftBits();
  if(charDone) {
    if(upTime > dahTime) {                               // If character finished (upTime > 3 ditTimes, aka dahTime)
      PrintCharacter();                                  // Print the character
      charDone = false;                                  // Reset flags for the next character
      charIndex = 0;
    }
    downTime = 0;                                        // Reset downTime timer
  }
}
//-------------------------------------------------------

void ShiftBits() {
  if(downTime < ditTime / 2) return;                     // Ignore if downTime too short (noise)

  charIndex = charIndex << 1;                            // Shift charIndex (points to character in LUT) left 1 bit (x2)

  if(downTime < midTime) {                               // downTime < midTime, read as a DIT
    charIndex++;                                         //   Add 1 to charIndex (increment)
    strcat(lcd0, ".");                                   //   Append dit mark to display string
    ditMin = min(ditMin, downTime);                      //   Set ditMin if downTime is less than current ditMin (for keying report)
    ditMax = max(ditMax, downTime);                      //   Set ditMax if downTime is greater than current ditMax (for keying report)
    ditAcc += downTime;                                  //   Add downTime to accumulated ditAcc time (for keying report)
    ditCount++;                                          //   Increment ditCount counter (for keying report)
  }

  else {                                                 // downTime > midTime, read as a DAH
                                                         //   add 0 to charIndex (no increment)
    strcat(lcd0, "-");                                   //   Append dah mark to display string
    dahMin = min(dahMin, downTime);                      //   Set dahMin if downTime is less than current dahMin (for keying report)
    dahMax = max(dahMax, downTime);                      //   Set dahMax if downTime is greater than current dahMax (for keying report)
    dahAcc += downTime;                                  //   Add downTime to accumulated dahAcc time (for keying report)
    dahCount++;                                          //   Increment dahCount counter (for keying report)
  }
  toneDone = false;
}
//-------------------------------------------------------

void PrintCharacter() {
  char xlatLUT[] = "^^TEMNAIOGKDWRUS^^QZYCXBJP^L^FVH09^8^^^7^^^^^^^61^^^^^^^2^^^3^45";    // Morse translation lookup table

  if(charIndex > 63) {                                   // Keyed character is punctuation
    PrintPunctuation();
  }

  else if(xlatLUT[charIndex] == '^') {                   // Keyed char is defined as punctuation in LUT
    PrintPunctuation();
  }

  else {
    sprintf(lcd1, "%c\0", xlatLUT[charIndex]);           // Keyed char is alpha-numeric, show char from LUT
    PrintLCD(lcd0, lcd1);                                // Print strings on LCD display and USB serial port
  }
  lcd0[0] = '\0';                                        // Clear lcd0 string
}
//-------------------------------------------------------

void PrintPunctuation() {
  switch(charIndex) {                                    // Punctuation and META chars requires additional description
    case 41:  strcpy(lcd1,"(  Left Paren  "); break;     // Left Paren -.--.
    case 42:  strcpy(lcd1,"   <Start>     "); break;     // META Start -.-.-
    case 45:  strcpy(lcd1,"/  Slash       "); break;     // Slash -..-.
    case 46:  strcpy(lcd1,"=  Equal       "); break;     // Equal/double dash -...-
    case 53:  strcpy(lcd1,"   <End>       "); break;     // META End .-.-.
    case 55:  strcpy(lcd1,"   <Wait>      "); break;     // META Wait .-...
    case 61:  strcpy(lcd1,"   <Understood>"); break;     // META Understood ...-.
    case 71:  strcpy(lcd1,":  Colon       "); break;     // Colon ---...
    case 76:  strcpy(lcd1,",  Comma       "); break;     // Comma --..--
    case 82:  strcpy(lcd1,")  Right Paren "); break;     // Right Paren -.--.-
    case 84:  strcpy(lcd1,"!  Exclamation "); break;     // Exclamation -.-.--
    case 85:  strcpy(lcd1,";  Semi-colon  "); break;     // Semicolon -.-.-.
    case 94:  strcpy(lcd1,"-  Dash        "); break;     // Dash -....-
    case 97:  strcpy(lcd1,"\'  Apostrophe "); break;     // Apostrophe .----.
    case 101: strcpy(lcd1,"@  At          "); break;     // At .--.-.
    case 106: strcpy(lcd1,".  Period      "); break;     // Period .-.-.-
    case 107: strcpy(lcd1,"   <Paragraph> "); break;     // META Paragraph .-.-..
    case 109: strcpy(lcd1,"\"  Quote      "); break;     // Quotation .-..-.
    case 114: strcpy(lcd1,"_  Underscore  "); break;     // Underline ..--.-
    case 115: strcpy(lcd1,"?  Question    "); break;     // Question ..--..
    case 122: strcpy(lcd1,"   <End Work>  "); break;     // META End of Work ...-.-
    case 246: strcpy(lcd1,"$  Dollar      "); break;     // Dollar ...-..-
    case 338: strcpy(lcd1,"   <CQ>        "); break;     // META CQ -.-.--.-
    case 511: strcpy(lcd1,"   <Error>     "); break;     // META Error ........
    default:  strcpy(lcd1," (?) Undefined "); break;     // Undefined
  }
  PrintLCD(lcd0, lcd1);                                  // Print strings on LCD display and USB serial port
}
//-------------------------------------------------------

// ####################################################
// ##  R E C E I V I N G   M O D E                   ##
// ####################################################
void RecvMode() {
  int c, i, d,
    anpgrp[3] = {27, 37, 52};                            // charLUT[0-27]Alpha only, [0-37]Alpha+numbers, [0-52]Alpha+Numbers+Punctuation

  strcpy(lcd0,"RECEIVING  MODE");
  if(!sound) {
    strcpy(lcd1, " Not available ");
    PrintLCD(lcd0, lcd1); delay(TWOSEC);
    return;
  }

  strcpy(lcd1,ver); 
  PrintLCD(lcd0, lcd1);                                  // Print to LCD display and USB serial port
  SoundCode(CQ);                                         // Sound CQ
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
  SetCharset();

  randomSeed(micros());                                  // Initialize random generator

  while(true) {
    d = random(1, 5) + 2;                                // Random select 3 - 7 characters
    strcpy(lcd0, "Listen carefull"); sprintf(lcd1, "to %i characters", d);
    PrintLCD(lcd0, lcd1); delay(ONESEC);

    lcd1[0] = '\0';                                      // Clear the lcd1 string
    for(i = 0; i < d; i++) {
      c = random(1, anpgrp[anp]);                        // Choose a random character from selected charset
      sprintf(lcd1, "%s%c ", lcd1, charLUT[c]);          // Append character to lcd1 string, but do not print yet
      SoundCode(c); delay(dahTime);                      // Sound morse code for chosen character
    }
    delay(FIVSEC);                                       // Pause 5 seconds for interpreting codes

    strcpy(lcd0, "Characters were");
    PrintLCD(lcd0, lcd1); delay(FIVSEC);                 // Print to LCD display and USB serial port

    strcpy(lcd0,""); strcpy(lcd1, "KEY now to exit");    // Show how to exit function and return to menu
    PrintLCD(lcd0, lcd1);                                // Delay(ONESEC);
    if(KeyDelay(ONESEC)) break;
  }
}
//-------------------------------------------------------

void SoundCode(int charIndex) {
  switch(charIndex) {
    case 0: break;                                       // Unassigned
    case 1: Dit(); Dah(); break;                         // A .-
    case 2: Dah(); Dit(); Dit(); Dit(); break;           // B -...
    case 3: Dah(); Dit(); Dah(); Dit(); break;           // C -.-.
    case 4: Dah(); Dit(); Dit(); break;                  // D -..
    case 5: Dit(); break;                                // E .
    case 6: Dit(); Dit(); Dah(); Dit(); break;           // F ..-.
    case 7: Dah(); Dah(); Dit(); break;                  // G --.
    case 8: Dit(); Dit(); Dit(); Dit(); break;           // H ....
    case 9: Dit(); Dit(); break;                         // I ..
    case 10: Dit(); Dah(); Dah(); Dah(); break;          // J .---
    case 11: Dah(); Dit(); Dah(); break;                 // K -.-
    case 12: Dit(); Dah(); Dit(); Dit(); break;          // L .-..
    case 13: Dah(); Dah(); break;                        // M --
    case 14: Dah(); Dit(); break;                        // N -.
    case 15: Dah(); Dah(); Dah(); break;                 // O ---
    case 16: Dit(); Dah(); Dah(); Dit(); break;          // P .--.
    case 17: Dah(); Dah(); Dit(); Dah(); break;          // Q --.-
    case 18: Dit(); Dah(); Dit(); break;                 // R .-.
    case 19: Dit(); Dit(); Dit(); break;                 // S ...
    case 20: Dah(); break;                               // T -
    case 21: Dit(); Dit(); Dah(); break;                 // U ..-
    case 22: Dit(); Dit(); Dit(); Dah(); break;          // V ...-
    case 23: Dit(); Dah(); Dah(); break;                 // W .--
    case 24: Dah(); Dit(); Dit(); Dah(); break;          // X -..-
    case 25: Dah(); Dit(); Dah(); Dah(); break;          // Y -.--
    case 26: Dah(); Dah(); Dit(); Dit(); break;          // Z --..
    case 27: Dah(); Dah(); Dah(); Dah(); Dah(); break;   // 0 -----
    case 28: Dit(); Dah(); Dah(); Dah(); Dah(); break;   // 1 .----
    case 29: Dit(); Dit(); Dah(); Dah(); Dah(); break;   // 2 ..---
    case 30: Dit(); Dit(); Dit(); Dah(); Dah(); break;   // 3 ...--
    case 31: Dit(); Dit(); Dit(); Dit(); Dah(); break;   // 4 ....-
    case 32: Dit(); Dit(); Dit(); Dit(); Dit(); break;   // 5 .....
    case 33: Dah(); Dit(); Dit(); Dit(); Dit(); break;   // 6 -....
    case 34: Dah(); Dah(); Dit(); Dit(); Dit(); break;   // 7 --...
    case 35: Dah(); Dah(); Dah(); Dit(); Dit(); break;   // 8 ---..
    case 36: Dah(); Dah(); Dah(); Dah(); Dit(); break;   // 9 ----.
    case 37: Dah(); Dit(); Dah(); Dah(); Dit(); break;   // ( -.--.
    case 38: Dah(); Dit(); Dit(); Dah(); Dit(); break;   // / -..-.
    case 39: Dah(); Dit(); Dit(); Dit(); Dah(); break;   // = -...-
    case 40: Dah(); Dah(); Dah(); Dit(); Dit(); Dit(); break;                           // : ---...
    case 41: Dah(); Dah(); Dit(); Dit(); Dah(); Dah(); break;                           // , --..--
    case 42: Dah(); Dit(); Dah(); Dah(); Dit(); Dah(); break;                           // ) -.--.-
    case 43: Dah(); Dit(); Dah(); Dit(); Dah(); Dah(); break;                           // ! -.-.--
    case 44: Dah(); Dit(); Dah(); Dit(); Dah(); Dit(); break;                           // ; -.-.-.
    case 45: Dah(); Dit(); Dit(); Dit(); Dit(); Dah(); break;                           // - -....-
    case 46: Dit(); Dah(); Dah(); Dah(); Dah(); Dit(); break;                           // ' .----.
    case 47: Dit(); Dah(); Dah(); Dit(); Dah(); Dit(); break;                           // @ .--.-.
    case 48: Dit(); Dah(); Dit(); Dah(); Dit(); Dah(); break;                           // . .-.-.-
    case 49: Dit(); Dit(); Dah(); Dah(); Dit(); Dah(); break;                           // _ ..--.-
    case 50: Dit(); Dit(); Dah(); Dah(); Dit(); Dit(); break;                           // ? ..--..
    case 51: Dit(); Dit(); Dit(); Dah(); Dit(); Dit(); Dah(); break;                    // $ ...-..-
    case CQ: Dah(); Dit(); Dah(); Dit(); Dah(); Dah(); Dit(); Dah(); break;             // CQ -.-.--.-
  }
}
//-------------------------------------------------------

// ####################################################
// ##  A N A L Y Z E   K E Y I N G                   ##
// ####################################################
void Analyze() {
  int avgDit = 0, avgDah = 0, wpmc;

  strcpy(lcd0, "KEYING ANALYSIS");
  strcpy(lcd1, ver);
  PrintLCD(lcd0, lcd1); delay(TWOSEC);                   // Print to LCD display and USB serial port
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

  if(ditCount > 0) avgDit = int(ditAcc / ditCount);      // Calculate average dit time
  sprintf(lcd0, "DIT (.)  < %i", midTime);
  sprintf(lcd1, "DIT Target %i", ditTime);
  PrintLCD(lcd0, lcd1); if(KeyDelay(FIVSEC)) Click();    // Print to LCD display and USB serial port

  sprintf(lcd0, "DIT Total  %i", ditAcc);
  sprintf(lcd1, "DIT Count  %i", ditCount);
  PrintLCD(lcd0, lcd1); if(KeyDelay(FIVSEC)) Click();    // Print to LCD display and USB serial port

  sprintf(lcd0, "DIT Target %i", ditTime);
  sprintf(lcd1, "DIT Avrage %i", avgDit);
  PrintLCD(lcd0, lcd1); if(KeyDelay(FIVSEC)) Click();    // Print to LCD display and USB serial port

  sprintf(lcd0, "DIT Max    %i", ditMax);
  sprintf(lcd1, "DIT Min    %i", ditMin);
  PrintLCD(lcd0, lcd1); if(KeyDelay(FIVSEC)) Click();    // Print to LCD display and USB serial port

  if(dahCount > 0) avgDah = int(dahAcc / dahCount);      // Calculate average dah time
  sprintf(lcd0, "DAH (-)  > %i", midTime);
  sprintf(lcd1, "DAH Target %i", dahTime);
  PrintLCD(lcd0, lcd1); if(KeyDelay(FIVSEC)) Click();    // Print to LCD display and USB serial port

  sprintf(lcd0, "DAH Total  %i", dahAcc);
  sprintf(lcd1, "DAH Count  %i", dahCount);
  PrintLCD(lcd0, lcd1); if(KeyDelay(FIVSEC)) Click();    // Print to LCD display and USB serial port

  sprintf(lcd0, "DAH Target %i", dahTime);
  sprintf(lcd1, "DAH Avrage %i", avgDah);
  PrintLCD(lcd0, lcd1); if(KeyDelay(FIVSEC)) Click();    // Print to LCD display and USB serial port

  sprintf(lcd0, "DAH Max    %i", dahMax);
  sprintf(lcd1, "DAH Min    %i", dahMin);
  PrintLCD(lcd0, lcd1); if(KeyDelay(FIVSEC)) Click();    // Print to LCD display and USB serial port

  wpmc = int(2400 / (avgDit + (avgDah/3)));              // Calculate WPM from measured averaged dit and dah times
  sprintf(lcd0, "WPM RATE   %i", wpm);
  sprintf(lcd1, "MEASURED   %i", wpmc);
  PrintLCD(lcd0, lcd1); if(KeyDelay(FIVSEC)) Click();    // Print to LCD display and USB serial port

  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

  strcpy(lcd0, " RESET KEY DATA");                       // Prompt to reset accumulated spec data
  strcpy(lcd1, "   KEY to OK   ");
  PrintLCD(lcd0, lcd1);                                  // Print to LCD display
  if(KeyDelay(TWOSEC)) {                                 // If KEY pressed in two seconds
    ditMin = dahMin = 9999;                              // Reset keying data to start values
    ditMax = ditAcc = ditCount = 0;
    dahMax = dahAcc = dahCount = 0;
    strcpy(lcd0, " Keying values ");
    strcpy(lcd1, "    CLEARED    ");
    PrintLCD(lcd0, lcd1); delay(FIVSEC);                 // Print to LCD display and USB serial port
  }
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
}
//-------------------------------------------------------
// ####################################################
// ##  T I M I N G   E X E R C I S E S               ##
// ####################################################
void Exercise() {                                        // Practice precise tone and space timings
  int indx = 1;

  strcpy(lcd0, " KEY EXERCISES ");
  strcpy(lcd1, ver);
  PrintLCD(lcd0, lcd1); delay(TWOSEC);                   // Print to LCD display and USB serial port
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

  strcpy(lcd0, "   EXERCISES   ");
  while(true) {                                          // Run menu loop until item selected
    switch(indx) {
      case 1: strcpy(lcd1, "   DIT  Tone   "); break;
      case 2: strcpy(lcd1, "   DAH  Tone   "); break;
      case 3: strcpy(lcd1, "IntraChar Pause"); break;
      case 4: strcpy(lcd1, "Character Pause"); break;
      case 5: strcpy(lcd1, "  Word  Pause  "); break;
      default: strcpy(lcd1, "     Cancel    "); break;
    }
    PrintLCD(lcd0, lcd1);
    if(KeyDelay(TWOSEC)) break;                          // Key pressed, exit from menu loop, pass indx to processing
    if(++indx > 5) indx = 0;                             // Otherwise increment indx and repeat menu loop
  }

  Click(); MarkLCD();                                    // Acknowledge selection
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
  lcd.write(LCD_CLR); delay(50);

  switch(indx) {                                         // Process selection per indx value
    case 1: KeyX(ditTime); break;                        // Practice DIT keypress at current WPM rate
    case 2: KeyX(dahTime); break;                        // Practice DAH keypress at current WPM rate
    case 3: PauseX(ditTime); break;                      // Practice intra character spaces (pause between DITs and DAHs)
    case 4: PauseX(3*ditTime); break;                    // Practice INTER character spaces (pause between characters)
    case 5: PauseX(7*ditTime); break;                    // Practice INTER word spaces (pause between words)
    default: break;                                      // Indx == 0 (Cancel) return to main menu
  }
}
//-------------------------------------------------------

void KeyX(unsigned int argt) {
  int x;

  strcpy(lcd0, "  KEY  TIMING  ");
  strcpy(lcd1, "Hold >2sec EXIT");
  PrintLCD(lcd0, lcd1); delay(TWOSEC);                   // Show function title
  sprintf(lcd1, "Target<> %u mS", argt);
  PrintLCD(lcd0, lcd1); delay(TWOSEC);                   // Show target time
  Dit();
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

  while(true) {
    while(digitalRead(PIN_KEY));                         // Wait until key press detected
    downStart = millis();
    sound ? tone(PIN_SPK, pitchLUT[pitch]) : DoNothing();

    while(!digitalRead(PIN_KEY));                        // Wait for key release
    downTime = millis() - downStart;                     // Calculate how long the key was down
    noTone(PIN_SPK);

    if(downTime > TWOSEC) break;                         // If key held down longer than 2 seconds

    if(argt >= downTime)
      x = min(7 + int(10 * (argt - downTime) / argt), 13);
    else
      x = max(7 - int(10 * (downTime - argt) / argt), 1); // Easier for handling negative results

    sprintf(lcd0, "<>=%u    *=%u", argt, downTime);
    strcpy(lcd1, "-     < >     +");                      // Preload blank graph string
    lcd1[x] = '*';                                        // Position result marker on graph
    PrintLCD(lcd0, lcd1);                                 // Show results
  }
  noTone(PIN_SPK);                                        // Stop audio tone
}
//-------------------------------------------------------

void PauseX(unsigned int argt) {                         // Measure pause between dits/dahs within a character
  int
    x;

  strcpy(lcd0, " PAUSE  TIMING ");
  strcpy(lcd1, "Open >2sec EXIT");
  PrintLCD(lcd0, lcd1); delay(TWOSEC);                   // Show function title

  sprintf(lcd1, "Target<> %u mS", argt);
  PrintLCD(lcd0, lcd1); delay(TWOSEC);                   // Show target time
  Dit(); Dit();
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

  while(true) {
    while(digitalRead(PIN_KEY));                         // Wait until key press detected
    sound ? tone(PIN_SPK, pitchLUT[pitch]) : DoNothing();

    while(!digitalRead(PIN_KEY));                        // Wait for key release
    upStart = millis();                                  // Record when key is released
    noTone(PIN_SPK);

    while(digitalRead(PIN_KEY));                         // Wait for key press
    upTime = millis() - upStart;                         // Calculate how long the key was up between tones
    sound ? tone(PIN_SPK, pitchLUT[pitch]) : DoNothing();

    while(!digitalRead(PIN_KEY));                        // Wait for key release
    noTone(PIN_SPK);

    if(upTime > TWOSEC) break;                           // If key open longer than 2 seconds between tones

    if(argt >= upTime)
      x = min(7 + int(10 * (argt - upTime) / argt), 13);
    else
      x = max(7 - int(10 * (upTime - argt) / argt), 1);

    sprintf(lcd0, "<>=%u    *=%u", argt, upTime);
    strcpy(lcd1, "-     < >     +");                      // Preload blank graph string
    lcd1[x] = '*';                                        // Position result marker on graph
    PrintLCD(lcd0, lcd1);                                 // Show results
  }
  noTone(PIN_SPK);                                        // Stop audio tone
}
//-------------------------------------------------------

// ####################################################
// ##  O P T I O N S   M E N U                       ##
// ####################################################
void OptionsMode() {
  int indx = 1;

  strcpy(lcd0, " OPTIONS  MENU ");
  strcpy(lcd1, ver);
  PrintLCD(lcd0, lcd1); delay(TWOSEC);
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

  strcpy(lcd0, " SELECT OPTION ");
  while(true) {                                          // Run menu loop until item selected
    switch(indx) {
      case 1: strcpy(lcd1, "Select WPM Rate"); break;
      case 2: strcpy(lcd1, "Select Char Set"); break;
      case 3: strcpy(lcd1, "Select CW Pitch"); break;
      case 4: strcpy(lcd1, "    Edit QRZ   "); break;
      default: strcpy(lcd1, "     Cancel    "); break;
    }
    PrintLCD(lcd0, lcd1);
    if(KeyDelay(TWOSEC)) break;                          // key pressed, exit from menu loop, pass indx to processing
    if(++indx > 4) indx = 0;                             // Otherwise increment indx and repeat menu loop
  }

  Click(); MarkLCD();                                    // Acknowledge selection
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
  lcd.write(LCD_CLR); delay(50);

  switch(indx) {                                         // Process function per indx value
    case 1: SetWPM(); break;
    case 2: SetCharset(); break;
    case 3: SetCWPitch(); break;
    case 4: SetQRZ(); break;
    default: break;                                      // Indx == 0 (Cancel) return to main menu
  }
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
}
//-------------------------------------------------------

void SetWPM() {
  int indx = wpm;
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

  strcpy(lcd0, "SELECT WPM RATE");
  while(true) {                                          // Run menu loop until item selected
    switch(indx) {
      case 5:  strcpy(lcd1, "      5 wpm    "); break;
      case 10: strcpy(lcd1, "     10 wpm    "); break;
      case 15: strcpy(lcd1, "     15 wpm    "); break;
      case 20: strcpy(lcd1, "     20 wpm    "); break;
      case 25: strcpy(lcd1, "     25 wpm    "); break;
      case 30: strcpy(lcd1, "     30 wpm    "); break;
      default: strcpy(lcd1, "     Cancel    "); break;   // indx = 0
    }
    PrintLCD(lcd0, lcd1);
    if(KeyDelay(TWOSEC)) break;                          // key pressed, exit from menu loop, pass indx to processing
    if((indx += 5) > 30) indx = 0;                       // Otherwise increment indx and repeat menu loop
  }
  
  Click(); MarkLCD();                                    // Acknowledge selection
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
  lcd.write(LCD_CLR); delay(50);

  if(indx > 0) {
    switch(indx) {                                         // Set wpm, etc values per indx value
      case 5: wpm =  5; ditTime = 240; midTime = 480; dahTime = 720; break;
      case 10: wpm = 10; ditTime = 120; midTime = 240; dahTime = 360; break;
      case 15: wpm = 15; ditTime = 80;  midTime = 160; dahTime = 240; break;
      case 20: wpm = 20; ditTime = 60;  midTime = 120; dahTime = 180; break;
      case 25: wpm = 25; ditTime = 48;  midTime =  96; dahTime = 144; break;
      case 30: wpm = 30; ditTime = 40;  midTime =  80; dahTime = 120; break;
      default: break;                                      // If indx is 0 (Cancel) do not change current values
    }
    while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

    strcpy(lcd1, "KEY now to save");
    PrintLCD(lcd0, lcd1);                                  // Print to LCD display
    if(KeyDelay(TWOSEC)) {
      EEPROM.put(EEP_WPM, wpm);                          // Save wpm, etc to EEPROM
      EEPROM.put(EEP_DIT, ditTime);                      // -- EEPROM.put does not save if value not changed
      EEPROM.put(EEP_MID, midTime);
      EEPROM.put(EEP_DAH, dahTime);
      sprintf(lcd0, "%i wpm", wpm);
      strcpy(lcd1, "Saved to EEPROM");
      PrintLCD(lcd0, lcd1); delay(FIVSEC);               // Print to LCD display and USB serial port
    }
  }
  else{                                                  // CANCEL selected
    sprintf(lcd0, "WPM rate %i", wpm);
    strcpy(lcd1, "CANCEL: NO save");
    PrintLCD(lcd0, lcd1); delay(FIVSEC);
  }
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
}
//-------------------------------------------------------

void SetCharset() {
  int indx = anp + 1;
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

  strcpy(lcd0, "SELECT CHAR SET");
  while(true) {                                          // Run menu loop until item selected
    switch(indx) {
      case 1: strcpy(lcd1, "  Alpha  Only  "); break;    // Alphabet characters only
      case 2: strcpy(lcd1, "  Alpha + Num  "); break;    // Alphabet and numbers
      case 3: strcpy(lcd1, "Alpha+Num+Punct"); break;    // Alphabet, numbers, and punctuations/macros
      default: strcpy(lcd1, "     Cancel    "); break;   // indx = 0
    }
    PrintLCD(lcd0, lcd1);
    if(KeyDelay(TWOSEC)) break;                          // Key pressed, exit from menu loop, pass indx to processing
    if(++indx > 3) indx = 0;                             // Otherwise increment indx and repeat loop
  }

  Click(); MarkLCD();                                    // Acknowledge selection
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
  lcd.write(LCD_CLR); delay(50);

  if(indx > 0) {
    anp = indx - 1;
    while(!digitalRead(PIN_KEY));                        // Ensure KEY is released before proceeding

    strcpy(lcd1, "KEY now to save");
    PrintLCD(lcd0, lcd1);

    if(KeyDelay(TWOSEC)) {                               // KEY detected, save anp to EEPROM
      EEPROM.put(EEP_ANP, anp);
      switch(anp) {                                      // Acknowledge saved
        case 0: strcpy(lcd0, "  Alpha Only   "); break;
        case 1: strcpy(lcd0, "  Alpha + Num  "); break;
        case 2: strcpy(lcd0, "Alpha+Num+Punct"); break;
      }
      strcpy(lcd1, "Saved to EEPROM");
      PrintLCD(lcd0, lcd1); delay(FIVSEC);
    }
  }
  else {                                                 // indx == 0, CANCEL selected, NO save
    switch(anp) {                                        // Acknowledge previous anp unchanged
      case 0: strcpy(lcd0, "  Alpha Only   "); break;
      case 1: strcpy(lcd0, "  Alpha + Num  "); break;
      case 2: strcpy(lcd0, "Alpha+Num+Punct"); break;
    }
    strcpy(lcd1, "CANCEL: NO save");
    PrintLCD(lcd0, lcd1); delay(FIVSEC);
  }
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
}
//-------------------------------------------------------

void SetCWPitch() {
  int indx = 1, tmp = pitch;
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
  strcpy(lcd0, "SELECT CW PITCH");

  if(!sound) {                                           // Function not available if sound disabled
    strcpy(lcd1, " Not available ");
    PrintLCD(lcd0, lcd1); delay(TWOSEC);
    return;
  }

  while(true) {                                          // Run menu loop until item selected
    switch(indx) {
      case 1: strcpy(lcd1, "     400 Hz    "); break;
      case 2: strcpy(lcd1, "     500 Hz    "); break;
      case 3: strcpy(lcd1, "     600 Hz    "); break;
      case 4: strcpy(lcd1, "     700 Hz    "); break;
      case 5: strcpy(lcd1, "     800 Hz    "); break;
      default: strcpy(lcd1, "     Cancel    "); break;   // indx = 0
    }
    PrintLCD(lcd0, lcd1);
    tone(PIN_SPK, pitchLUT[indx - 1], 500);              // Sound tone for 0.5 second
    if(KeyDelay(TWOSEC)) break;                          // Key pressed, exit from menu loop, pass indx to processing
    if(++indx > 5) indx = 0;                             // Otherwise increment indx and repeat loop
  }

  Click(); MarkLCD();                                    // Acknowledge selection
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
  lcd.write(LCD_CLR); delay(50);

  if(indx > 0){
    pitch = indx - 1;
    while(!digitalRead(PIN_KEY));                        // Ensure KEY is released before proceeding

    strcpy(lcd1, "KEY now to save");
    PrintLCD(lcd0, lcd1);
    if(KeyDelay(TWOSEC)) {
      EEPROM.put(EEP_PIT, pitch);                        // Save pitch to EEPROM
      sprintf(lcd0,"CW PITCH %i Hz", pitchLUT[pitch]);
      strcpy(lcd1, "Saved to EEPROM");
      PrintLCD(lcd0, lcd1); delay(FIVSEC);
    }
  }
  else {                                                 // CANCEL selected, do not save to EEPROM
    pitch = tmp;
    sprintf(lcd0, "CW PITCH %i", wpm);
    strcpy(lcd1, "CANCEL: NO save");
    PrintLCD(lcd0, lcd1); delay(FIVSEC);
  }
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
}
//-------------------------------------------------------

void SetQRZ() {                                          // Edit qrz to custom call sign or other string, 15 chars max
  int chrPos, chrVal;
  strcpy(lcd0, "    EDIT QRZ   "); strcpy(lcd1, qrz);
  PrintLCD(lcd0, lcd1); delay(TWOSEC);
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

  strcpy(lcd0, qrz);
  strcpy(lcd1, "               ");
  PrintLCD(lcd0, lcd1);

  lcd.write(LCD_OC);                                     // Turn cursor ON
  chrPos = 0;

  while(chrPos < LCD_COL) {                              // Loop each character until end of display reached
    lcd.write(LCD_RO1 + chrPos); lcd.write('^');         // Mark the current active character position
    chrVal = 0;                                          // Start secondary with first letter (space)

    while(true) {                                        // Loop through each available character
      lcd.write(LCD_RO0 + chrPos);                       // Place cursor at active character position
      chrVal == 0 ? lcd.write('_') : lcd.write(charLUT[chrVal]);  // Show '_' for space, otherwise show character
      Click();

      if(KeyDelay(750)) {                                // Key pressed, save current character to buffer string
        lcd0[chrPos] = charLUT[chrVal];
        Click();
        while(!digitalRead(PIN_KEY));                    // Ensure KEY is released before proceeding
        break;
      }
      if(++chrVal > 36) chrVal = 0;                      // Loop from '9' (36) back to ' ' (0)
    }
    lcd.write(LCD_RO1 + chrPos); lcd.write(' ');         // Clear marker ^
    chrPos++;                                            // Increment p to next character, and repeat loop
  }

  lcd.write(LCD_ON);                                     // Turn cursor OFF
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding

  strcpy(lcd1, "KEY now to save");
  PrintLCD(lcd0, lcd1);

  if(KeyDelay(TWOSEC)) {
      strcpy(qrz, lcd0);
      EEPROM.put(EEP_QRZ, qrz);                          // Save qrz to EEPROM
      strcpy(lcd1, "Saved to EEPROM");
      PrintLCD(lcd0, lcd1); delay(FIVSEC);
  }
  else {
    EEPROM.get(EEP_QRZ, qrz);                            // NOT SAVED, restore previous saved qrz
    strcpy(lcd1, "CANCEL: NO save");
    PrintLCD(lcd0, lcd1); delay(FIVSEC);
  }
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
}
//-------------------------------------------------------

// ####################################################
// ##  S U P P O R T   F U N C T I O N S             ##
// ####################################################
void EepromInit() {                                      // Load all default variables into EEPROM to format memory assignments
  EEPROM.put(EEP_QRZ, qrz);                              //   qrz[16] = "     K7DUN     ") Default pre-loaded at powerup
  EEPROM.put(EEP_WPM, wpm);                              //   15 WPM Default pre-loaded at powerup
  EEPROM.put(EEP_DIT, ditTime);                          //   standard dit time at 15 WPM Default pre-loaded at powerup
  EEPROM.put(EEP_MID, midTime);                          //   standard mid time at 15 WPM (= 2*dit) Default pre-loaded at powerup
  EEPROM.put(EEP_DAH, dahTime);                          //   standard dah time at 15 WPM (= 3*dit) Default pre-loaded at powerup
  EEPROM.put(EEP_ANP, anp);                              //   anp = 0 (Alpha only) Default pre-loaded at powerup
  EEPROM.put(EEP_PIT, pitch);                            //   note = 3 Default pre-loaded at powerup
  strcpy(lcd0," Format EEPROM ");
  strcpy(lcd1,"     READY     ");
  PrintLCD(lcd0, lcd1);
  Click();
  while(!digitalRead(PIN_KEY));                          // Ensure KEY is released before proceeding
}
//-------------------------------------------------------

void EepromRead() {                                      // Read values from EEPROM into respective variables
  EEPROM.get(EEP_QRZ, qrz);                              //   qrz <- default or as saved in prior session
  EEPROM.get(EEP_WPM, wpm);                              //   wpm <- default or as saved in prior session
  EEPROM.get(EEP_DIT, ditTime);                          //   ditTime <- default or as saved in prior session
  EEPROM.get(EEP_MID, midTime);                          //   midTime <- default or as saved in prior session
  EEPROM.get(EEP_DAH, dahTime);                          //   dahTime <- default or as saved in prior session
  EEPROM.get(EEP_ANP, anp);                              //   anp <- defai;t pr as saved in prior session
  EEPROM.get(EEP_PIT, pitch);                            //   note <- default or as saved in prior session
}
//-------------------------------------------------------

void SoundQRZ() {                                        // Sound qrz in morse at current wpm rate
  int i = 0;

  do {
    SoundCode(int(strchr(charLUT, qrz[i]) - charLUT));   // Sound the morse equivelent for each character in string qrz
    delay(ditTime * 2);                                  // Delay 2 additional dit times for 3 dit INTERchar pause
  } while(qrz[++i] != '\0');                             // Repeat until end of string qrz is found
}
//-------------------------------------------------------

void Dit() {                                             // Sound tone for 1 ditTime (set by WPM rate)
  sound ? tone(PIN_SPK, pitchLUT[pitch], ditTime) : DoNothing();
  delay(ditTime * 2);                                    // Delay for dit sound + 1 dit intrachar pause
}
//-------------------------------------------------------

void Dah() {                                             // Sound tone for 1 dahTime (set by WPM rate)
  sound ? tone(PIN_SPK, pitchLUT[pitch], dahTime) : DoNothing();
  delay(ditTime * 4);                                    // Delay for dah sound + 1 dit intrachar pause
}
//-------------------------------------------------------

void Click() {                                           // "Click" sound for audio acknowledgement
  tone(PIN_SPK, 1000, 10);
  delay(ditTime);
}
//-------------------------------------------------------

void PrintLCD(char* sout0, char* sout1) {                // Print passed string pointers to LCD display and USB serial port
  lcd.write(LCD_CLR); delay(50);                         // Output to LCD display
  lcd.write(LCD_RO0); lcd.write(sout0);
  lcd.write(LCD_RO1); lcd.write(sout1);

  Serial.write("-----\n");                               // Output to USB serial port
  Serial.write(sout0); Serial.write('\t');
  Serial.write(sout1); Serial.write('\n');
}
//-------------------------------------------------------

void MarkLCD() {                                         // Place marker at right edge of lower lcd line. Does not affect buffer strings
  lcd.write(0xA3); lcd.write('*');
}
//-------------------------------------------------------

bool KeyDelay(int tlimit) {                              // Delay with key press interrupt
  long t;

  lcd.write(0x8F); lcd.write('*');                       // Display visual prompt KeyDelay is active
  t = millis();

  while((millis() - t) < tlimit) {                       // You have tlimit milliseconds to press KEY
    if(!digitalRead(PIN_KEY)) {                          // If keypress detected ...
      lcd.write(0x8F); lcd.write(' ');                   //   Erase visual prompt
      return(true);                                      //   Return true
    }
  }
                                                         // If tlimit timed out without keypress ...
  lcd.write(0x8F); lcd.write(' ');                       //   Erase visual prompt
  return(false);                                         //   Return false
}
//-------------------------------------------------------

void DoNothing() {
  return;                                                // A do-nothing function to use where needed
}

//-------------------------------------------------------

// ####################################################
// ##  E N D   O F   C O D E                         ##
// ####################################################
